use base64::engine::general_purpose::STANDARD as BASE64_STANDARD;
use base64::Engine;
use std::collections::HashMap;

use crate::Res;

#[derive(Debug, Clone, serde::Deserialize)]
pub struct OpenIDConfig {
    pub authorization_endpoint: String,
    pub token_endpoint: String,
    pub userinfo_endpoint: String,
}

#[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
pub struct TokenResponse {
    pub access_token: String,
    pub token_type: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub refresh_token: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub expires_in: Option<u64>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub id_token: Option<String>,
}

#[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
pub struct UserInfo {
    pub sub: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub name: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub given_name: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub family_name: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub preferred_username: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub email: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub email_verified: Option<bool>,
}

impl OpenIDConfig {
    /// Load OpenID configuration from a given URL
    pub async fn load_from(url: &str) -> Res<Self> {
        Ok(reqwest::get(url).await?.json().await?)
    }

    /// Get the authorization URL where a user should be redirect
    pub fn authorization_url(&self, client_id: &str, state: &str, redirect_uri: &str) -> String {
        let client_id = urlencoding::encode(client_id);
        let state = urlencoding::encode(state);
        let redirect_uri = urlencoding::encode(redirect_uri);

        format!("{}?response_type=code&scope=openid%20profile%20email&client_id={client_id}&state={state}&redirect_uri={redirect_uri}", self.authorization_endpoint)
    }

    /// Query the token endpoint
    pub async fn request_token(
        &self,
        client_id: &str,
        client_secret: &str,
        code: &str,
        redirect_uri: &str,
    ) -> Res<(TokenResponse, String)> {
        let authorization = BASE64_STANDARD.encode(format!("{}:{}", client_id, client_secret));

        let mut params = HashMap::new();
        params.insert("grant_type", "authorization_code");
        params.insert("code", code);
        params.insert("redirect_uri", redirect_uri);

        let response = reqwest::Client::new()
            .post(&self.token_endpoint)
            .header("Authorization", format!("Basic {authorization}"))
            .form(&params)
            .send()
            .await?
            .text()
            .await?;

        Ok((serde_json::from_str(&response)?, response))
    }

    /// Query the UserInfo endpoint
    pub async fn request_user_info(&self, token: &TokenResponse) -> Res<(UserInfo, String)> {
        let response = reqwest::Client::new()
            .get(&self.userinfo_endpoint)
            .header("Authorization", format!("Bearer {}", token.access_token))
            .send()
            .await?
            .text()
            .await?;

        Ok((serde_json::from_str(&response)?, response))
    }
}
